import struct
import numpy as np
from .var_int import get_length_of_varint, parse_varint


def read_short(f):
    ''' Read a signed 2 byte integer from `f`'''
    return struct.unpack('<h', f.read(2))[0]


def read_unsigned_short(f):
    ''' Read an unsigned 2 byte integer from `f`'''
    return struct.unpack('<H', f.read(2))[0]


def read_int(f):
    ''' Read a signed 4 byte integer from `f`'''
    return struct.unpack('<i', f.read(4))[0]


def read_unsigned_int(f):
    ''' Read an signed 4 byte integer from `f`'''
    return struct.unpack('<I', f.read(4))[0]


def read_float(f):
    ''' Read a 4 byte float from `f`'''
    return struct.unpack('<f', f.read(4))[0]


def read_double(f):
    ''' Read an 8 byte float from `f`'''
    return struct.unpack('<d', f.read(8))[0]


def read_array(f, dtype, count):
    '''Read a numpy array with `dtype` of length `count` from file-like `f`'''
    dt = np.dtype(dtype)
    if count == 0:
        return np.array((), dtype=dtype)
    return np.frombuffer(f.read(count * dt.itemsize), count=count, dtype=dt)


def read_string(f):
    '''Read a string from eventio file or object f.
    Eventio stores strings as a short giving the length
    of the string and the string itself.
    '''
    length = read_short(f)
    return f.read(length)


def read_var_string(f):
    '''Read a string from eventio file or object f.
    This string is similar to the one in `read_string` but uses
    the variable length integer instead of a short.
    '''
    length = read_unsigned_varint(f)
    return f.read(length)


def read_from(f, fmt):
    '''
    read the struct fmt specification from file f
    Moves the current position.
    '''
    result = struct.unpack_from(
        fmt,
        f.read(struct.calcsize(fmt))
    )
    return result


def read_ints(f, n_ints):
    ''' read n ints from file f '''
    return read_from(f, '{:d}i'.format(n_ints))


def read_time(f):
    '''Read a time as combination of seconds and nanoseconds'''
    sec, nano = read_from(f, '<ii')
    return sec, nano


def read_varint(f):
    u = read_unsigned_varint(f)
    # u values of 0,1,2,3,4,... here correspond to signed values of
    #   0,-1,1,-2,2,... We have to test the least significant bit:
    if (u & 1) == 1:  # Negative number;
        return -(u >> 1) - 1
    else:
        return u >> 1


def read_unsigned_varint(f):
    '''this returns a python integer'''
    var_int_bytes = bytearray(f.read(1))
    var_int_length = get_length_of_varint(var_int_bytes[0])
    if var_int_length > 1:
        var_int_bytes += f.read(var_int_length - 1)

    return parse_varint(var_int_bytes)



import numpy as np
from .base import EventIOObject
from .version_handling import assert_version_in
from .tools import read_short, read_int, read_string, read_float, read_array


class Histograms(EventIOObject):
    eventio_type = 100

    def parse(self):
        assert_version_in(self, [1, 2])
        self.seek(0)

        n_histograms = read_short(self)

        histograms = []
        for i in range(n_histograms):
            hist = {}
            hist['type'] = self.read(1).decode('ascii')
            hist['title'] = read_string(self).decode('utf-8')
            if len(hist['title']) % 2 == 0:
                self.read(1)
            hist['id'] = read_int(self)
            hist['n_bins_x'] = read_short(self)
            hist['n_bins_y'] = read_short(self)
            hist['entries'] = read_int(self)
            hist['tentries'] = read_int(self)

            if hist['n_bins_y'] > 0:
                axes = 'xy'
                n_counts = hist['n_bins_x'] * hist['n_bins_y']
            else:
                axes = 'x'
                n_counts = hist['n_bins_x']

            for ax in axes:
                hist['underflow_' + ax] = read_int(self)
                hist['overflow_' + ax] = read_int(self)

                if hist['type'] in {'R', 'r', 'F', 'D'}:
                    hist['lower_' + ax] = read_float(self)
                    hist['upper_' + ax] = read_float(self)
                    hist['sum_' + ax] = read_float(self)
                    hist['tsum_' + ax] = read_float(self)
                else:
                    hist['lower_' + ax] = read_int(self)
                    hist['upper_' + ax] = read_int(self)
                    hist['sum_' + ax] = read_int(self)
                    hist['tsum_' + ax] = read_int(self)

            if hist['type'] in 'FD':
                hist['content_all'] = read_float(self)
                hist['content_inside'] = read_float(self)
                hist['content_outside'] = read_array(self, dtype='<f4', count=8)

                if hist['tentries'] > 0:
                    hist['data'] = read_array(self, dtype='<f4', count=n_counts)
                else:
                    hist['data'] = np.zeros(n_counts)
            else:
                if hist['tentries'] > 0:
                    hist['data'] = read_array(self, dtype='<i4', counts=n_counts)
                else:
                    hist['data'] = np.zeros(n_counts)

            if hist['n_bins_y'] > 0:
                hist['data'] = hist['data'].reshape((hist['n_bins_y'], hist['n_bins_x']))

            histograms.append(hist)

        return histograms



import numpy as np
from functools import lru_cache


dt1 = np.dtype([
    # (fieldname, type, shape)
    ('setup_id', 'i4'),
    ('trigger_mode', 'i4'),
    ('min_pixel_mult', 'i4'),
    ('n_pixels', 'i4'),
])


@lru_cache()
def build_dt2(n_pixels):
    return np.dtype([
        ('pixel_HV_DAC', 'i4', (n_pixels,)),
        ('n_drawers', 'i4'),
    ])


@lru_cache()
def build_dt3(version, n_drawers):
    dt = [
        ('threshold_DAC', 'i4', (n_drawers,)),
        ('n_drawers', 'i4'),
        ('ADC_start', 'i2', (n_drawers,)),
        ('ADC_count', 'i2', (n_drawers,)),
    ]

    if version >= 1:
        dt.extend([
            ('time_slice', 'f4'),
            ('sum_bins', 'i4'),
        ])

    return np.dtype(dt)


# nref_shape = tools.get_scount(data)
# lref_shape = tools.get_scount(data)
def build_dt4(n_ref_shape, l_ref_shape):
    return np.dtype([
        ('ref_step', 'f4'),
        ('ref_shape', 'f2', (n_ref_shape, l_ref_shape)),
    ])
